"use strict";
// MIT License
//
// Copyright (c) 2021 Augustin Husson
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
Object.defineProperty(exports, "__esModule", { value: true });
exports.Complete = exports.analyzeCompletion = exports.ContextKind = exports.operatorTerms = exports.matcherTerms = void 0;
const tree_1 = require("./tree");
const language_1 = require("@codemirror/language");
const parser_terms_1 = require("../grammar/parser.terms");
const path_finder_1 = require("../parser/path-finder");
exports.matcherTerms = [{ label: '!=' }, { label: '=~' }, { label: '=' }, { label: '>' }, { label: '>=' }, { label: '<' }, { label: '<=' }];
exports.operatorTerms = [{ label: 'OR' }, { label: 'AND' }];
// ContextKind is the different possible value determinate by the autocompletion
var ContextKind;
(function (ContextKind) {
    ContextKind[ContextKind["KeyPath"] = 0] = "KeyPath";
    ContextKind[ContextKind["Pattern"] = 1] = "Pattern";
    ContextKind[ContextKind["QueryMatcher"] = 2] = "QueryMatcher";
    ContextKind[ContextKind["QueryOperator"] = 3] = "QueryOperator";
})(ContextKind = exports.ContextKind || (exports.ContextKind = {}));
function arrayToCompletionResult(data, from, to, span = true) {
    return {
        from: from,
        to: to,
        options: data,
        span: span ? /^[a-zA-Z0-9_:]+$/ : undefined,
    };
}
function computeStartCompletePosition(node, pos) {
    let start = node.from;
    if (node.type.id === parser_terms_1.QueryPath) {
        start = pos;
    }
    else if (node.type.id === parser_terms_1.KVSearch || node.type.id === parser_terms_1.Expression || node.type.id === parser_terms_1.QueryNode) {
        start = pos;
    }
    return start;
}
function calculateQueryPath(state, node, pos) {
    const terms = (0, path_finder_1.retrieveAllRecursiveNodes)((0, path_finder_1.walkBackward)(node, parser_terms_1.Query), parser_terms_1.QueryPath, parser_terms_1.Identifier, parser_terms_1.Regexp);
    const decodedTerms = [];
    let depth = 0;
    let i = 0;
    for (const term of terms) {
        if (term.type.id === parser_terms_1.Identifier) {
            decodedTerms.push(state.sliceDoc(term.from, term.to));
        }
        else {
            decodedTerms.push(new RegExp(state.sliceDoc(term.from, term.to)));
        }
        if (node.type.id === parser_terms_1.Identifier && term.from === node.from && term.to === node.to) {
            depth = i;
        }
        else if (node.type.id === parser_terms_1.QueryPath && ((term.type.id === parser_terms_1.Identifier && term.to === pos - 1) ||
            // it's pos -2 when it's a Regexp, because a Regexp is identified between `//` so it required one more char to access to the end of the Regexp expression.
            (term.type.id === parser_terms_1.Regexp && term.to === pos - 2))) {
            // This case is quite particular. We are in this situation:
            // `labels.` where the cursor is after the dot.
            // The current node is a `QueryPath` and the tree looks like this: KVSearch(Expression(Query(QueryPath(Identifier,⚠),⚠))).
            // In this case there is basically no node that is matching the position of the cursor.
            // So an idea is to actually matched the first Identifier that is one position before (so in this case the Identifier matching `labels`).
            // And since we are at the next level in the tree we need to increase the depth to one point.
            depth = i + 1;
        }
        else if (node.type.id === parser_terms_1.QueryPath && term.type.id === parser_terms_1.Regexp && term.to === pos - 1) {
            // this is a special case when typically you finished to write your regexp but you didn't start to type the next path.
            // For example : labels./e.*/ where the cursor is after the second '/'. It shouldn't autocomplete anything.
            return null;
        }
        i++;
    }
    return { terms: decodedTerms, depth: depth };
}
function getCloserErrorNodeFromPosition(node, pos) {
    const cursor = node.cursor;
    const candidates = [];
    while (cursor.next()) {
        // Note: 0 is the id of the error node.
        if (cursor.from === cursor.to && cursor.node.type.id === 0) {
            candidates.push(cursor.node);
        }
    }
    // let's loop other the candidates and let's find the closest node from our position.
    let result = null;
    for (const candidate of candidates) {
        if (result === null || candidate.to - pos < result.to - pos) {
            result = candidate;
        }
    }
    return result;
}
function analyzeRootNode(state, node, pos) {
    const result = [];
    const errorNode = getCloserErrorNodeFromPosition(node, pos);
    if (errorNode === null) {
        // in this case, it means we have a correct expression and so the only thing that can be completed are the query operators
        result.push({ kind: ContextKind.QueryOperator, treeTerms: { terms: [], depth: 0 } });
        return result;
    }
    const parent = errorNode.parent;
    if (parent === null) {
        return result;
    }
    switch (parent.type.id) {
        case parser_terms_1.QueryNode:
        case parser_terms_1.Expression:
        case parser_terms_1.KVSearch:
            // we are likely at the beginning of a new expression so we can safely autocomplete the QueryPath.
            result.push({ kind: ContextKind.KeyPath, treeTerms: { terms: [], depth: 0 } });
            break;
        case parser_terms_1.Query:
            if ((0, path_finder_1.containsAtLeastOneChild)(parent, parser_terms_1.Neq, parser_terms_1.EqlRegex, parser_terms_1.EqlSingle)) {
                // we likely have this expression:
                // labels.env !=
                // So we should autocomplete the QueryPattern
                // eslint-disable-next-line no-case-declarations
                const treeTerms = calculateQueryPath(state, errorNode, pos);
                if (treeTerms !== null) {
                    result.push({
                        kind: ContextKind.Pattern,
                        treeTerms: { terms: treeTerms.terms, depth: treeTerms.terms.length }
                    });
                }
                break;
            }
            else {
                result.push({ kind: ContextKind.QueryMatcher, treeTerms: { terms: [], depth: 0 } });
            }
    }
    return result;
}
function analyzeQueryPattern(state, node, pos, result) {
    const treeTerms = calculateQueryPath(state, node, pos);
    if (treeTerms !== null) {
        // here we are autocompleting the pattern that should be associated to the query path. So we want the value of the last node in the tree.
        // That's why we are changing the depth here instead of using the one calculated.
        result.push({
            kind: ContextKind.Pattern,
            treeTerms: { terms: treeTerms.terms, depth: treeTerms.terms.length }
        });
    }
}
function analyzeQueryPath(state, node, pos, result) {
    const treeTerms = calculateQueryPath(state, node, pos);
    if (treeTerms !== null) {
        result.push({
            kind: ContextKind.KeyPath,
            treeTerms: treeTerms
        });
    }
}
function analyzeCompletion(state, node, pos) {
    let result = [];
    switch (node.type.id) {
        case parser_terms_1.Expression:
        case parser_terms_1.KVSearch:
            result = result.concat(analyzeRootNode(state, node, pos));
            break;
        case parser_terms_1.Pattern:
            analyzeQueryPattern(state, node, pos, result);
            break;
        case parser_terms_1.Identifier:
        case parser_terms_1.QueryPath:
            // this is the usual case when user is typing the path/list of key that should be used for the search.
            // Like `labels.instance`.
            // Here we have to know what is currently requested to be autocompleted (with the example above: labels or instance).
            // In a later stage it will be used to give the position in the tres and change the list to complete.
            analyzeQueryPath(state, node, pos, result);
            break;
    }
    return result;
}
exports.analyzeCompletion = analyzeCompletion;
class Complete {
    constructor(objects) {
        if (objects !== undefined) {
            this.tree = (0, tree_1.newRootNode)(objects);
            this.objectList = objects;
        }
    }
    kvSearch(context) {
        const { state, pos } = context;
        const tree = (0, language_1.syntaxTree)(state).resolve(pos, -1);
        const contexts = analyzeCompletion(state, tree, pos);
        let result = [];
        for (const context of contexts) {
            switch (context.kind) {
                case ContextKind.KeyPath:
                    result = this.autocompleteQueryPath(result, context);
                    break;
                case ContextKind.Pattern:
                    result = this.autocompleteQueryPattern(result, context);
                    break;
                case ContextKind.QueryMatcher:
                    result = result.concat(exports.matcherTerms);
                    break;
                case ContextKind.QueryOperator:
                    result = result.concat(exports.operatorTerms);
                    break;
            }
        }
        return arrayToCompletionResult(result, computeStartCompletePosition(tree, pos), pos);
    }
    autocompleteQueryPath(result, context) {
        if (this.objectList === undefined || this.tree === undefined) {
            return result;
        }
        const node = (0, tree_1.iterateAndCreateMissingChild)(context.treeTerms.depth, context.treeTerms.terms, this.objectList, this.tree);
        if (node !== null) {
            return result.concat(node.keys.map((value) => ({ label: value, type: 'text' })));
        }
        else {
            return result;
        }
    }
    autocompleteQueryPattern(result, context) {
        if (this.objectList === undefined || this.tree === undefined) {
            return result;
        }
        const node = (0, tree_1.iterateAndCreateMissingChild)(context.treeTerms.depth, context.treeTerms.terms, this.objectList, this.tree);
        if (node !== null) {
            return result.concat(node.values.map((value) => ({ label: value, type: 'text' })));
        }
        else {
            return result;
        }
    }
}
exports.Complete = Complete;
