"use strict";
// MIT License
//
// Copyright (c) 2021 Augustin Husson
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
Object.defineProperty(exports, "__esModule", { value: true });
exports.translate = void 0;
const parser_terms_1 = require("../grammar/parser.terms");
const path_finder_1 = require("../parser/path-finder");
const language_1 = require("@codemirror/language");
function buildQuery(state, query) {
    // first let's calculate the queryPath
    const keyPath = [];
    const terms = (0, path_finder_1.retrieveAllRecursiveNodes)(query, parser_terms_1.QueryPath, parser_terms_1.Identifier, parser_terms_1.Regexp);
    for (const term of terms) {
        if (term.type.id === parser_terms_1.Identifier) {
            keyPath.push(state.sliceDoc(term.from, term.to));
        }
        else {
            keyPath.push(new RegExp(state.sliceDoc(term.from, term.to)));
        }
    }
    let match;
    if (query.getChild(parser_terms_1.Neq) !== null) {
        match = 'negative';
    }
    else if (query.getChild(parser_terms_1.EqlSingle) !== null) {
        match = 'exact';
    }
    else if (query.getChild(parser_terms_1.EqlRegex) !== null) {
        match = 'fuzzy';
    }
    else if (query.getChild(parser_terms_1.Gtr) !== null) {
        match = 'greater';
    }
    else if (query.getChild(parser_terms_1.Gte) !== null) {
        match = 'greaterEqual';
    }
    else if (query.getChild(parser_terms_1.Lss) !== null) {
        match = 'less';
    }
    else if (query.getChild(parser_terms_1.Lte) !== null) {
        match = 'lessEqual';
    }
    else {
        return null;
    }
    let pattern = '';
    const patternNode = query.getChild(parser_terms_1.Pattern);
    if (patternNode !== null) {
        pattern = state.sliceDoc(patternNode.from, patternNode.to);
    }
    else {
        return null;
    }
    return {
        keyPath: keyPath,
        match: match,
        pattern: pattern,
    };
}
function buildQueryOperator(queryNode) {
    if (queryNode.getChild(parser_terms_1.And)) {
        return 'and';
    }
    else {
        return 'or';
    }
}
// TODO to be improved by transforming it with an iteration or with a better recursive function
function translateRec(state, node) {
    if (node === null) {
        return null;
    }
    switch (node.type.id) {
        case parser_terms_1.KVSearch:
        case parser_terms_1.Expression:
            return translateRec(state, node.firstChild);
        case parser_terms_1.Query:
            return buildQuery(state, node);
        case parser_terms_1.QueryNode:
            // eslint-disable-next-line no-case-declarations
            const left = translateRec(state, node.firstChild);
            // eslint-disable-next-line no-case-declarations
            const right = translateRec(state, node.lastChild);
            // eslint-disable-next-line no-case-declarations
            const operator = buildQueryOperator(node);
            return {
                operator: operator,
                left: left,
                right: right,
            };
    }
    return null;
}
function translate(state) {
    const root = (0, language_1.syntaxTree)(state).topNode;
    return translateRec(state, root);
}
exports.translate = translate;
